import decimal
from functools import partial
from typing import List

from .. import types
from ..method import method


@method('dna_identity')
def get_identity(address: str) -> types.Identity: pass

@method('dna_identities')
def get_identities() -> List[types.Identity]: pass

@method('dna_state')
def get_current_process() -> types.State: pass

@method('dna_getCoinbaseAddr')
def get_coinbase_address() -> str: pass

@method('dna_getBalance')
def get_balance(address: str) -> types.Balance: pass

@method('dna_sendTransaction')
def send_tx(
    nonce:          int             = None,
    epoch:          int             = None,
    from_:          str             = None,
    to:             str             = None,
    amount:         decimal.Decimal = None,
    maxFee:         decimal.Decimal = None,
    payload:        str             = None,
    tips:           decimal.Decimal = None,
    useProto:       bool            = None,
    type:           types.TxType   = types.TxType.SendTx,
) -> str: pass


@method('dna_sendInvite')
def send_invite(
    nonce:          int             = None,
    epoch:          int             = None,
    to:             str             = None,
    amount:         decimal.Decimal = None,
) -> types.Invite: pass

@method('dna_activateInvite')
def activate_invite(
    nonce:          int             = None,
    epoch:          int             = None,
    key:            str             = None,
) -> types.Invite: pass

kill_identity = partial(send_tx, type=types.TxType.KillTx)

kill_invitee = partial(send_tx, type=types.TxType.KillInviteeTx)

send_dna = partial(send_tx, type=types.TxType.SendTx)

change_god_address = partial(send_tx, type=types.TxType.ChangeGodAddressTx)

@method('dna_becomeOnline')
def become_online(
    nonce:          int             = None,
    epoch:          int             = None,
) -> str: pass

@method('dna_becomeOffline')
def become_offline(
    nonce:          int             = None,
    epoch:          int             = None,
) -> str: pass

@method('dna_epoch')
def get_epoch() -> types.Epoch: pass

@method('dna_ceremonyIntervals')
def get_ceremony_intervals() -> types.CeremonyIntervals: pass

@method('dna_exportKey')
def export_key(password: str) -> str: pass

@method('dna_burn')
def burn(
    nonce:          int             = None,
    epoch:          int             = None,
    from_:          str             = None,
    amount:         decimal.Decimal = None,
    maxFee:         decimal.Decimal = None,
    key:            str             = None,
) -> str: pass

@method('dna_changeProfile')
def change_profile(
    info:           str             = None,
    nickname:       str             = None,
    maxFee:         decimal.Decimal = None,
) -> types.ChangeProfile: pass

@method('dna_profile')
def get_profile(
    address:        str             = None,
) -> types.Profile: pass

@method('dna_activateInviteToRandAddr')
def activate_invite_to_random_address(
    nonce:          int             = None,
    epoch:          int             = None,
    key:            str             = None,
) -> types.ActivateInviteToRandomAddr: pass

@method('dna_isValidationReady')
def isValidationReady() -> bool: pass





import decimal
from typing import List, Any

from .. import types
from ..method import method


@method('contract_getStake')
def get_stake(contract: str) -> types.Stake: pass

@method('contract_estimateDeploy')
def estimate_deploy(
    from_:      str                     = None,
    codeHash:   str                     = None,
    amount:     decimal.Decimal         = None,
    maxFee:     decimal.Decimal         = None,
    args:       List[types.DynamicArg]  = None,
) -> types.ContractTxReceipt: pass

@method('contract_estimateCall')
def estimate_call(
    from_:          str                     = None,
    contract:       str                     = None,
    method:         str                     = None,
    amount:         decimal.Decimal         = None,
    maxFee:         decimal.Decimal         = None,
    args:           List[types.DynamicArg]  = None,
    broadcastBlock: int                     = None,
) -> types.ContractTxReceipt: pass

@method('contract_estimateTerminate')
def estimate_terminate(
    from_:          str                     = None,
    contract:       str                     = None,
    maxFee:         decimal.Decimal         = None,
    args:           List[types.DynamicArg]  = None,
) -> types.ContractTxReceipt: pass

@method('contract_deploy')
def deploy_contract(
    from_:      str                     = None,
    codeHash:   str                     = None,
    amount:     decimal.Decimal         = None,
    maxFee:     decimal.Decimal         = None,
    args:       List[types.DynamicArg]  = None,
) -> str: pass

@method('contract_call')
def call_contract(
    from_:          str                     = None,
    contract:       str                     = None,
    method:         str                     = None,
    amount:         decimal.Decimal         = None,
    maxFee:         decimal.Decimal         = None,
    args:           List[types.DynamicArg]  = None,
    broadcastBlock: int                     = None,
) -> str: pass

@method('contract_terminate')
def terminate_contract(
    from_:          str                     = None,
    contract:       str                     = None,
    maxFee:         decimal.Decimal         = None,
    args:           List[types.DynamicArg]  = None,
) -> str: pass

@method('contract_readData')
def read_contract_data(
    contract:   str = None,
    key:        str = None,
    format:     str = None,
) -> Any: pass

@method('contract_readonlyCall')
def readonly_call_contract(
    contract:   str                    = None,
    method:     str                    = None,
    format:     str                    = None,
    args:       List[types.DynamicArg] = None,
) -> Any: pass

@method('contract_subscribeToEvent')
def subscribe_to_contract_event(
    contract:   str = None,
    event:      str = None,
) -> None: pass

@method('contract_unsubscribeFromEvent')
def unsubscribe_from_contract_event(
    contract:   str = None,
    event:      str = None,
) -> None: pass

@method('contract_events')
def read_events(
    contract:   str = None,
) -> None: pass





from typing import List

from ..method import method


@method('account_list')
def get_accounts() -> List[str]: pass

@method('account_create')
def create_account(password: None) -> str: pass

@method('account_unlock')
def unlock_account(
    address:    str,
    password:   str = None,
    time:       int = None
) -> None: pass

@method('account_lock')
def lock_account(address: str) -> None: pass





from typing import List, Tuple

from .. import types
from ..method import method


@method('flip_submit')
def submit_flip(
    publicHex:  str = None,
    privateHex: str = None,
    pairId:     str = None,
    hex:        str = None
) -> types.FlipSubmit: pass

@method('flip_shorthashes')
def get_short_flip_hashes() -> List[types.FlipHashes]: pass

@method('flip_longHashes')
def get_long_flip_hashes() -> List[types.FlipHashes]: pass

@method('flip_get')
def get_flip(hash: str) -> types.Flip: pass

@method('flip_getRaw')
def get_raw_flip(hash: str) -> types.RawFlip: pass

@method('flip_words')
def flip_words(hash: str) -> Tuple[int, int]: pass

@method('flip_delete')
def delete_flip(hash: str) -> str: pass

@method('flip_submitShortAnswers')
def submit_short_answers(answers: types.FlipAnswers) -> str: pass

@method('flip_submitLongAnswers')
def submit_long_answers(answers: types.FlipAnswers) -> str: pass




import decimal
from typing import Optional

from .. import types
from ..method import method


@method('bcn_lastBlock')
def get_last_block() -> types.Block: pass

@method('bcn_blockAt')
def get_block_by_height(height: int) -> types.Block: pass

@method('bcn_block')
def get_block_by_hash(hash: str) -> types.Block: pass

@method('bcn_transaction')
def get_transaction(hash: str) -> types.Transaction: pass

@method('bcn_mempool')
def get_mempool() -> Optional[str]: pass

@method('bcn_syncing')
def check_sync() -> types.Sync: pass

@method('bcn_transactions')
def get_address_transactions(address: str, count: int = 5, token: str = None) -> types.AddressTransactions: pass

@method('bcn_pendingTransactions')
def get_address_pending_transactions(address: str) -> types.AddressTransactions: pass

@method('bcn_burntCoins')
def get_burnt_coins() -> Optional[types.BurntCoins]: pass

@method('bcn_sendRawTx')
def send_raw_tx(tx: str) -> str: pass

@method('bcn_getRawTx')
def get_raw_tx(
    nonce:          int             = None,
    epoch:          int             = None,
    from_:          str             = None,
    to:             str             = None,
    amount:         decimal.Decimal = None,
    maxFee:         decimal.Decimal = None,
    payload:        str             = None,
    tips:           decimal.Decimal = None,
    useProto:       bool            = None,
    type:           types.TxType   = types.TxType.SendTx,
) -> str: pass

@method('bcn_feePerByte')
def get_fee_rate() -> decimal.Decimal: pass

@method('bcn_txReceipt')
def get_transaction_receipt(hash: str) -> Optional[types.TxReceipt]: pass





class IdenaException(Exception):
    def __init__(self, code, message):
        self.code = code
        self.message = message

    def __repr__(self):
        return f'Idena RPC Client Exception: {self.code} - {self.message}'



import decimal
from datetime import datetime
from typing import List, Optional, NamedTuple


class Block(NamedTuple):
    coinbase:       str
    hash:           str
    parentHash:     str
    height:         int
    timestamp:      int
    root:           str
    identityRoot:   str
    ipfsCid:        str
    transactions:   List[str]
    flags:          List[str]
    isEmpty:        bool
    offlineAddress: str

class Transaction(NamedTuple):
    hash:           str
    type:           str
    from_:          str
    to:             Optional[str]
    amount:         str
    tips:           str
    maxFee:         str
    nonce:          int
    epoch:          int
    payload:        str
    blockHash:      str
    usedFee:        str
    timestamp:      int

class Sync(NamedTuple):
    syncing:        bool
    currentBlock:   int
    highestBlock:   int
    wrongTime:      bool
    genesisBlock:   int

class AddressTransactions(NamedTuple):
    token:          Optional[str]
    transactions:   Optional[List[Transaction]]

class BurntCoins(NamedTuple):
    address:        str
    amount:         decimal.Decimal
    key:            str

class BaseTxArgs(NamedTuple):
    nonce:          int             = None
    epoch:          int             = None

class TxType:
    SendTx                  = 0
    ActivationTx            = 1
    InviteTx                = 2
    KillTx                  = 3
    SubmitFlipTx            = 4
    SubmitAnswersHashTx     = 5
    SubmitShortAnswersTx    = 6
    SubmitLongAnswersTx     = 7
    EvidenceTx              = 8
    OnlineStatusTx          = 9
    KillInviteeTx           = 10
    ChangeGodAddressTx      = 11
    BurnTx                  = 12
    ChangeProfileTx         = 13

class SendTxArgs(BaseTxArgs):
    from_:          str             = None
    to:             str             = None
    amount:         decimal.Decimal = None
    maxFee:         decimal.Decimal = None
    payload:        str             = None
    tips:           decimal.Decimal = None
    useProto:       bool            = None
    type:           TxType          = TxType.SendTx

class TxEvent(NamedTuple):
    eventName: str
    data:      List[bytes]

class TxAddr(NamedTuple):
    TxHash:     str
    Address:    str

class TxReceipt(NamedTuple):
    contractAddress: str
    success:         bool
    gasUsed:         int
    gasCost:         int
    from_:           str
    txHash:          str
    error:           str
    events:          List[TxEvent]

class FlipWords(NamedTuple):
    words:  List[int]
    used:   bool
    id:     int

class Identity(NamedTuple):
    address:                str
    profileHash:            str
    stake:                  decimal.Decimal
    invites:                int
    age:                    int
    state:                  str
    pubkey:                 str
    requiredFlips:          int
    availableFlips:         int
    flipKeyWordPairs:       Optional[List[FlipWords]]
    madeFlips:              int
    totalQualifiedFlips:    int
    totalShortFlipPoints:   float
    flips:                  Optional[List[str]]
    online:                 bool
    generation:             int
    code:                   bytes
    invitees:               Optional[List[TxAddr]]
    penalty:                decimal.Decimal
    lastValidationFlags:    Optional[List[str]]

class State(NamedTuple):
    name:   str

class Balance(NamedTuple):
    stake:   decimal.Decimal
    balance: decimal.Decimal
    nonce:   int

class Invite(NamedTuple):
    hash:       str
    receiver:   str
    key:        str

class Epoch(NamedTuple):
    epoch:                  int
    nextValidation:         datetime
    currentPeriod:          str
    currentValidationStart: datetime

class CeremonyIntervals(NamedTuple):
    FlipLotteryDuration:  float
    ShortSessionDuration: float
    LongSessionDuration:  float

class ChangeProfile(NamedTuple):
    txHash: str
    hash:   str

FlipSubmit = ChangeProfile

class Profile(NamedTuple):
    info:       bytes
    nickname:   str

class ActivateInviteToRandomAddr(NamedTuple):
    hash:       str
    address:    str
    key:        str

class FlipHashes(NamedTuple):
    hash:      str
    ready:     bool
    extra:     bool
    available: bool

class Flip(NamedTuple):
    hex:        str
    privateHex: str

class RawFlip(NamedTuple):
    publicHex:  str
    privateHex: str

class Answer(NamedTuple):
    none  = '0'
    left  = '1'
    right = '2'

class Grade:
    gradeNone      = '0'
    gradeReported  = '1'
    gradeD         = '2'
    gradeC         = '3'
    gradeB         = '4'
    gradeA         = '5'

class FlipAnswer(NamedTuple):
    hash:       str
    grade:      Grade
    answer:     Answer
    wrongWords: List[bool]

class FlipAnswers(NamedTuple):
    answers:  List[FlipAnswer]

class Peer(NamedTuple):
    id:     str
    addr:   str

class Stake(NamedTuple):
    Hash: str
    Stake: decimal.Decimal

class DynamicArg(NamedTuple):
    index:  int
    format: str
    value:  str

class ContractTxReceipt:
    contract:   str
    success:    bool
    gasUsed:    int
    gasCost:    decimal.Decimal
    txHash:     str
    txFee:      decimal.Decimal
    error:      str





import functools
import os
from inspect import signature
from typing import Any, get_type_hints

import requests

from .exceptions import IdenaException


req_id: int = 0
RPC_NODE: str = os.environ.get('IDENA_RPC_NODE')
API_KEY: str = os.environ.get('IDENA_API_KEY')


# TODO: theese methods differ from other single arg methods
# the other methods accept a single item list as param
# but this one need an object
SPECIAL_FUNCS = [
    'get_raw_tx', 'get_address_pending_transactions',
    'send_dna', 'send_invite', 'change_profile',
    'activate_invite_to_random_address',
    'flip_submit', 'submit_short_answers',
    'submit_long_answers', 'estimate_deploy',
    'estimate_call', 'estimate_terminate',
    'deploy_contract', 'call_contract', 'terminate_contract',
    'readonly_call_contract', 'subscribe_to_contract_event',
    'unsubscribe_from_contract_event', 'read_events',
]

def parse_params(f, *args, **kwargs):
    bind_params = signature(f).bind(*args, **kwargs)
    bind_params.apply_defaults()
    bind_args = bind_params.arguments
    params = {}

    for k,v in bind_args.items():
        if v is None:
            continue

        if k == 'from_':
            params['from'] = v
        else:
            params[k] = v

    if len(params) == 1 and f.__name__ not in SPECIAL_FUNCS:
        return list(params.values())

    elif len(params) == 0:

        # Becuase of strange behaviour of this api
        if f.__name__ in SPECIAL_FUNCS:
            return [{}]

        return []

    return [dict(params)]


def parse_result(data: dict, type_: Any):
    if data is None:
        return

    if type_ in [bool, int, float, str, bytes]:
        try:
            return type_(data)
        except TypeError:
            return type_(data[0])

    if 'from' in data:
        data['from_'] = data['from']
        del data['from']

    if isinstance(data, str):
        return type_(data)

    if isinstance(data, list) and len(type_.__args__) == 1:
        return [type_.__args__[0](**o) for o in data]

    if len(data) == 1:
        return type_(*data)

    return type_(**data)


def method(method_name: str):
    def method_factory(func):
        return_type = get_type_hints(func)['return']

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            params = parse_params(func, *args, **kwargs)
            data = dict(
                key=API_KEY,
                id=req_id,
                method=method_name,
                params=params,
            )
            result = requests.post(RPC_NODE, json=data).json()
            if 'error' in result:
                raise IdenaException(
                    result['error']['code'],
                    result['error']['message'],
                )

            return parse_result(result['result'], return_type)
        return wrapper
    return method_factory




from typing import List

from .. import types
from ..method import method


@method('net_peers')
def get_peers() -> List[types.Peer]: pass

@method('net_addPeer')
def add_peers(url) -> str: pass

@method('net_ipfsAddress')
def get_ipfs_addr() -> str: pass





# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['idena', 'idena.apis']

package_data = \
{'': ['*']}

install_requires = \
['requests>=2.25.0,<3.0.0']

setup_kwargs = {
    'name': 'idena',
    'version': '0.2.2',
    'description': 'Idena RPC client',
    'long_description': ".. Idena Client documentation master file, created by\n   sphinx-quickstart on Wed Dec  2 12:50:15 2020.\n   You can adapt this file completely to your liking, but it should at least\n   contain the root `toctree` directive.\n\nIdena Client\n============\n\n.. image:: https://travis-ci.com/iRhonin/idena.py.svg?branch=main\n    :target: https://travis-ci.com/iRhonin/idena.py\n\nThis is Python wrapper for Idene Node.\n\n* Highly inspired by `Web3.py <https://github.com/ethereum/web3.py>`_ and `cookiecutter-pypackage <https://github.com/briggySmalls/cookiecutter-pypackage>`_.\n\nInstallation\n------------\n\nIdena.py can be installed using ``pip`` as follows:\n\n.. code-block:: shell\n\n   $ pip install idena\n\nFor the development, clone the repository then:\n\n.. code-block:: shell\n\n   $ poetry install\n\n\nUsing Idena\n-----------\n\nThis library depends on a connection to an Idena node and there are 2 ways to configure them. \n\nCalling `client.init` \n*********************\n\n.. code-block:: python\n\n   >>> from idena import client\n   >>> client.init('http://localhost:9009/', 'api-key')\n\n\nSetting environment variables\n*****************************\n\nSet `IDENA_RPC_NODE` and `IDENA_API_KEY` envars:\n\n.. code-block:: sh\n\n   $ export IDENA_RPC_NODE=http://localhost:9009/\n   $ export IDENA_API_KEY=api-key\n\nGetting Blockchain Info\n-----------------------\n\n.. code-block:: python\n   \n   >>> client.blockchain.get_last_block()\n      \n      Block(coinbase='0xbe854231db69ab042073b7ff8309ae3ee265a40f', \n         hash='0xa88e6ab305d7ee311ad2de35338cdbf7e664d860709e5a53f0307baeeaa6f968', \n         parentHash='0xe324a208892241e0294e5a6334965660375dda7a3ad8d8a42a5f3f2ef2857a22', \n         height=2159398, \n         timestamp=1606904853, \n         root='0xd874709bdd4c6fcd95e2e531cc07a4ce42ab23334dfd12ceb45350535b36664c', \n         identityRoot='0x9f3661f19e13d4860f2f2f1610abbbaf86abc8adf1a2781b371189e683745a97', \n         ipfsCid=None, \n         transactions=None, \n         flags=['OfflinePropose'], \n         isEmpty=False, \n         offlineAddress='0x0df427ad7e1906ab4fcc5fd31118932256f5dc7a')",
    'author': 'Arash Fatahzade',
    'author_email': 'fatahzade@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/iRhonin/idena.py',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from .apis import \
    account, \
    blockchain, \
    contract, \
    dna, \
    flip, \
    net


def init(rpc_node: str, api_key: str) -> None:
    """Initialization

    Args:
        rpc_node (str): URL of rpc node
        api_key (str): API-Key
    """
    from . import method

    method.RPC_NODE = rpc_node
    method.API_KEY = api_key



